#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Assortment of utility scripts implementing recipes from Python cookbooks, such as iterators, sorters, range queries, etc. 
"""

from jcvi.apps.base import dmain


if __name__ == '__main__':
    dmain(__file__)
